from exposer.exposer import Exposer
from exposer.actions import PathAction
from exposer.targets import Wsgi
from exposer.actions import CliAction
from exposer.targets import Cli
from wsgiref.simple_server import make_server


class app(object):
    cli = Exposer(action=CliAction, target=Cli)
    web = Exposer(action=PathAction, target=Wsgi)


class people:
    @staticmethod
    @app.web
    @app.cli
    def greetings(
        name=None, surname=None, greeting="Hello"
    ):
        print(
            "{}, {} {}".format(
                greeting,
                name.capitalize(),
                surname.capitalize(),
            )
        )


@app.web
@app.cli
def index():
    print("Welcome")


@app.cli
def run():
    APP = app.web.expose()
    app.web.rename("/index", "/")
    httpd = make_server("localhost", 8051, APP)
    import threading

    t = threading.Thread(None, httpd.serve_forever)
    t.start()
    import time

    while True:
        try:
            time.sleep(2)
        except KeyboardInterrupt as ki:
            httpd.shutdown()
            t.join()
            break


app.cli.expose()



from std import std
import sys
from parsers import Functionality, Parameter, Input
from formatters import (
    Output,
    Error,
    Documentation,
    Introspection,
)


class Exposable(object):
    starts_with = "eX://"
    auth_splitter = "@"
    method_seperator = "-"
    function_pointer = "!"
    instance_pointer = "#"
    ends_with = "/"

    def __init__(
        self,
        functionality=Functionality,
        parameter=Parameter,
        input=Input,
        output=Output,
        error=Error,
        introspection=Introspection,
        documentation=Documentation,
    ):
        self.input_set = (sys.argv, sys.stdin)
        self.functionality = functionality
        self.parameter = parameter
        self.input = input
        self.output = output
        self.error = error
        self.introspection = introspection

    def __call__(self, *args, **kwargs):
        self.function = self.functionality.parse(
            self.input_set
        )

        args, kwargs = self.parameter.parse(self.input_set)
        input = self.input.parse(self.input_set)

        with std(stdin_file=input, stderr_file="log") as s:
            result = self.function(*args, *kwargs)

        output = self.output.format(s.stdout.getvalue())
        if output:
            print(output)

        error = self.error.format(s.stderr.getvalue())
        if error:
            raise error

        return result

    def __repr__(self):
        hsh = ""
        if hasattr(self.function, "__qualname__"):
            hsh = (
                str(self.function_pointer)
                .join(
                    self.function.__qualname__.split(".")
                )
                .lower()
            )

        if hasattr(self.function, "__self__"):
            tmp = hsh.split(str(self.function_pointer))
            method_list = (
                str(self.method_seperator).join(tmp[0:-1]),
            )
            funcname = str(tmp[-1])
            hsh = (
                str(self.function_pointer).join(
                    [method_list, funcname]
                )
                + str(self.instance_pointer)
                + str(self.function.__self__)
            )
        hsh = (
            str(self.starts_with)
            + hsh
            + str(self.ends_with)
        )
        return hsh



import io
import sys
import threading


class Inp(io.StringIO):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self, *args, **kwargs):
        super().read(self, *args, **kwargs)

    def write(self, *args, **kwargs):
        super().write(*args, **kwargs)


class Out(io.StringIO):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self, *args, **kwargs):
        super().read(self, *args, **kwargs)

    def write(self, *args, **kwargs):
        print(*args, **kwargs, file=sys.__stdout__, end="")
        super().write(*args, **kwargs)


class Err(io.StringIO):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self, *args, **kwargs):
        super().read(self, *args, **kwargs)

    def write(self, *args, **kwargs):
        print(*args, **kwargs, file=sys.__stderr__, end="")
        super().write(*args, **kwargs)


class Environment(object):
    def __init__(
        self,
        stdin=sys.stdin,
        stdout=sys.stdout,
        stderr=sys.stderr,
    ):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr


class std(object):
    rlock = threading.RLock()

    def __init__(
        self,
        stdin_file=None,
        stdout_file=None,
        stderr_file=None,
    ):
        self.stdin = Inp()
        self.stdout = Out()
        self.stderr = Err()
        self.old_stdin = sys.stdin
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr

    def __enter__(self):
        if std.rlock.acquire():
            sys.stdin = self.stdin
            sys.stdout = self.stdout
            sys.stderr = self.stderr
            return Environment(
                self.stdin, self.stdout, self.stderr
            )

    def __exit__(self, *args):
        sys.stdin = self.old_stdin
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        std.rlock.release()


with std() as s:
    print("hello")
    print("world")

# print(s.stdout.getvalue())



class Terminal256(object):
    def __init__(self, functionality):
        self.exposable = functionality

    def parse_uri(self, string) -> (Function, Params):
        pass

    def take_input(self, fp):
        pass

    def print_output(self, string):
        pass

    def print_error(self, string):
        pass
